@@ -52,6 +52,7 @@ G_DEFINE_TYPE (NMIODINEPlugin, nm_iodine_plugin, NM_TYPE_VPN_PLUGIN)
 
 typedef struct {
 	GPid pid;
+	NMVPNPluginFailure failure;
 	GHashTable *ip4config;
 } NMIODINEPluginPrivate;
 
@@ -271,13 +272,23 @@ value_destroy (gpointer data)
 	g_slice_free (GValue, val);
 }
 
-static gboolean
-iodine_parse_stderr_line (const char* line, GHashTable *ip4config)
+static gint
+iodine_parse_stderr_line (NMVPNPlugin *plugin,
+						  const char* line,
+						  GHashTable *ip4config)
 {
-	gchar **split;
+	NMIODINEPluginPrivate *priv = NM_IODINE_PLUGIN_GET_PRIVATE (plugin);
+	gchar **split = NULL;
 	GValue *val;
 	gint len;
-	gint ret = TRUE;
+	gint ret = 1;
+
+	if (g_str_has_prefix(line, "Bad password")) {
+		g_debug ("Login failure");
+		priv->failure = NM_VPN_PLUGIN_FAILURE_LOGIN_FAILED;
+		ret = -1;
+		goto out;
+	}
 
 	split = g_strsplit (line, " ", 0);
 	len = g_strv_length (split);
@@ -340,7 +351,7 @@ iodine_parse_stderr_line (const char* line, GHashTable *ip4config)
 		g_hash_table_insert (ip4config,
 							 NM_VPN_PLUGIN_IP4_CONFIG_PREFIX,
 							 val);
-		ret = FALSE;
+		ret = 0; /* success */
 	} else
 		g_debug("%s", line);
 
@@ -369,7 +380,7 @@ iodine_stderr_cb (GIOChannel *source, GIOCondition condition, gpointer plugin)
 	if (l)
 		line[l-1] = '\0';
 
-	ret = iodine_parse_stderr_line(line, priv->ip4config);
+	ret = iodine_parse_stderr_line(plugin, line, priv->ip4config);
 	if (!ret) {
 		g_debug("Parsing done, sending IP4 config");
 		nm_vpn_plugin_set_ip4_config(plugin, priv->ip4config);
@@ -406,8 +417,10 @@ iodine_watch_cb (GPid pid, gint status, gpointer user_data)
 	waitpid (priv->pid, NULL, WNOHANG);
 	priv->pid = 0;
 
-	/* Must be after data->state is set since signals use data->state */
-	if (error) {
+	if (priv->failure >= 0) {
+		nm_vpn_plugin_failure (NM_VPN_PLUGIN (plugin),
+							   priv->failure);
+	} else if (error) {
 		nm_vpn_plugin_failure (NM_VPN_PLUGIN (plugin),
 							   NM_VPN_PLUGIN_FAILURE_CONNECT_FAILED);
 	}
@@ -422,6 +435,22 @@ has_user(const char* user)
 	return (getpwnam(user) == NULL) ? FALSE : TRUE;
 }
 
+
+static void
+send_password(gint fd, NMSettingVPN *s_vpn)
+{
+	const char *passwd;
+
+	passwd = nm_setting_vpn_get_secret (s_vpn, NM_IODINE_KEY_PASSWORD);
+	/* Don't send an empty password since this makes iodine block */
+	if (!passwd || !strlen(passwd))
+		passwd = "<none>";
+
+	write (fd, passwd, strlen(passwd));
+	write (fd, "\n", 1);
+}
+
+
 static gint
 nm_iodine_start_iodine_binary(NMIODINEPlugin *plugin,
 										 NMSettingVPN *s_vpn,
@@ -433,7 +462,7 @@ nm_iodine_start_iodine_binary(NMIODINEPlugin *plugin,
 	GSource *iodine_watch;
 	GIOChannel *stderr_channel;
 	gint	stdin_fd, stderr_fd;
-	const char *props_topdomain, *props_fragsize, *props_nameserver, *passwd;
+	const char *props_topdomain, *props_fragsize, *props_nameserver;
 
 	/* Find iodine */
 	iodine_binary = iodine_binary_paths;
@@ -458,12 +487,6 @@ nm_iodine_start_iodine_binary(NMIODINEPlugin *plugin,
 													 NM_IODINE_KEY_NAMESERVER);
 	props_topdomain = nm_setting_vpn_get_data_item (s_vpn,
 													NM_IODINE_KEY_TOPDOMAIN);
-
-	passwd = nm_setting_vpn_get_secret (s_vpn, NM_IODINE_KEY_PASSWORD);
-
-	if (passwd && strlen(passwd))
-		g_setenv("IODINE_PASS", passwd, TRUE);
-
 	iodine_argv = g_ptr_array_new ();
 	g_ptr_array_add (iodine_argv, (gpointer) (*iodine_binary));
 	/* Run in foreground */
@@ -505,7 +528,9 @@ nm_iodine_start_iodine_binary(NMIODINEPlugin *plugin,
 	g_ptr_array_free (iodine_argv, TRUE);
 
 	g_message ("iodine started with pid %d", pid);
-	close(stdin_fd);
+
+	send_password (stdin_fd, s_vpn);
+	close (stdin_fd);
 
 	stderr_channel = g_io_channel_unix_new (stderr_fd);
 	g_io_add_watch(stderr_channel,
@@ -531,7 +556,7 @@ real_connect (NMVPNPlugin   *plugin,
               GError       **error)
 {
 	NMSettingVPN *s_vpn;
-	gint iodine_fd = -1;
+	gint ret = -1;
 
 	s_vpn = NM_SETTING_VPN (nm_connection_get_setting (connection,
 													   NM_TYPE_SETTING_VPN));
@@ -542,9 +567,9 @@ real_connect (NMVPNPlugin   *plugin,
 	if (!nm_iodine_secrets_validate (s_vpn, error))
 		goto out;
 
-	iodine_fd = nm_iodine_start_iodine_binary (NM_IODINE_PLUGIN (plugin),
+	ret = nm_iodine_start_iodine_binary (NM_IODINE_PLUGIN (plugin),
 											   s_vpn, error);
-	if (!iodine_fd)
+	if (!ret)
 		return TRUE;
 
  out:
@@ -622,6 +647,7 @@ nm_iodine_plugin_init (NMIODINEPlugin *plugin)
 											 g_str_equal,
 											 NULL,
 											 value_destroy);
+	priv->failure = -1;
 }
 
 static void
